package etc.caradmin.mvc.controller.charging;

import etc.caradmin.io.sql.SQLFactory;
import etc.io.ClassFactory;
import lombok.extern.log4j.Log4j2;
import org.json.JSONArray;
import org.json.JSONObject;
import org.restlet.ext.json.JsonRepresentation;
import org.restlet.resource.Post;
import org.restlet.resource.ServerResource;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

//hmac
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.SignatureException;
import java.util.Formatter;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

/**
 * EnodeWebhookApi er en Serverlet router klasse som tar imot API kall fra Enode.
 *
 * @author Sindre Emil Vikre
 */
@Log4j2
public class EnodeWebhookAPI extends ServerResource {

    private static final String key = "0Svb4tDYQ34COMBiWlt5XJSoppXMhWDiVfetWaEm";
    public static final String environment = "sandbox";

    private static final String HMAC_SHA1_ALGORITHM = "HmacSHA1";

    public String getEnvironment(){
        return environment;
    }

    /**
     * getWebhook tar imot webhooken sendt fra Enode
     *
     * @param representation
     */
    @Post("application/json")
    public void getWebhook(final JsonRepresentation representation) {


        try {
            JSONArray jsonArr = representation.getJsonArray(); //Henter json array

            if(verifyEnodeSignature(jsonArr)) { //Verifiserer at kallet kommer fra Enode

                //Itererer gjennom json array og finner events med info om at en ny lading er registrert.
                var objList = findReleventEvent(jsonArr);

                if(!objList.isEmpty()) {
                    findChargingData(objList); //Henter ladinger hvis listen med events ikke er tom.
                }

            } else {
                System.out.println("Kall til /EnodeWebhookAPI kunne ikke autentiseres");
            }

        }catch(Exception e){
            System.out.println("Invalid json object from body.");
        }
    }

    /**
     *
     * @param payload
     * @return - bool som forteller om kallet ble verifisert eller ikke.
     */
    private boolean verifyEnodeSignature(JSONArray payload) {

        //Henter signatur fra header
        var enodeSignature = getRequest().getHeaders().getFirstValue("x-enode-signature");

        String hmac = null;
        try {

            //Genererer HMAC
            hmac = calculateRFC2104HMAC(payload.toString(), key);
           
        } catch(SignatureException | NoSuchAlgorithmException | InvalidKeyException e) {
            e.printStackTrace();
            System.out.println("Could not create hmac");
        }

        hmac = "sha1=" + hmac;

        return hmac.equals(enodeSignature); //returnerer true eller false
    }

    /**
     * findReleventEvent leter gjennom json array fra Enode notifikasjon og
     * finner ut om vi skal hente lade info eller ikke. Hvis en tom liste blir
     * returnert skal man ikke hente ladinger fra Enode.
     *
     * @param jsonArr
     * @return
     */
    public List<JSONObject> findReleventEvent(JSONArray jsonArr){

        List<JSONObject> objList = new ArrayList<JSONObject>();

        try {
            //Lager en liste med objekter som har gjeldende event
            for(int i = 0; i < jsonArr.length(); i++) {
                JSONObject obj = jsonArr.getJSONObject(i);
                String event = obj.getString("event");

                if(Objects.equals(event, "user:charger:updated")) {
                    JSONArray updated = obj.getJSONArray("updatedFields");


                    boolean isPluggedInUpdated = false;

                    for(int k = 0; k < updated.length(); k++) {
                        String updatedField = updated.getString(k);
                        if(Objects.equals(updatedField, "chargeStateIsPluggedIn")) {
                            isPluggedInUpdated = true;
                            k += updated.length() + 1; //Stopper lokken om isPluggedIn blir true
                        }
                    }

                    if(isPluggedInUpdated) {

                        boolean charger = obj.getJSONObject("charger").getJSONObject("chargeState").getBoolean("isPluggedIn");

                        if(!charger) { //Hvis laderen har blitt plugget ut.
                            objList.add(obj);
                        }
                    }
                }
            }

            return objList;
        }catch(Exception e){
            System.out.println("Problem when filtering events in EnodeWebhookAPI");
        }

        objList.clear(); //Make sure that list is empty if something went wrong.
        return objList;
    }

    /**
     *
     * @param objList
     */
    private void findChargingData(List<JSONObject> objList) {


        GetChargingDataEnode getChargingObj = new GetChargingDataEnode();

        String accessToken = getChargingObj.getAccesToken();

        if(accessToken != null){
            for(JSONObject obj : objList) {

                try {
                    //Henter frste object i json array
                    //JSONObject obj = jsonArr.getJSONObject(i);
                    //Lader objekt
                    JSONObject chargerObj = obj.getJSONObject("charger");

                    //Id til laderen
                    String chargerId = chargerObj.getString("id");

                    JSONObject chargerInfo = chargerObj.getJSONObject("information");

                    //Merket til laderen
                    String brand = chargerInfo.getString("brand");

                    //Objekt med status til laderen
                    JSONObject chargeStateObj = chargerObj.getJSONObject("chargeState");
                    //Henter dato til sist gang laderstatus med oppdatert
                    String date = chargeStateObj.getString("lastUpdated");
                    date = date.substring(0, date.length() - 5);
                    date = date.replace("T", " ");

                    JSONObject userInfo = obj.getJSONObject("user");
                    String customerSolution = userInfo.getString("id");

                    String jsonStr = getChargingObj.getChargingData(customerSolution, chargerId, date, accessToken);


                    String query = makeQuery(jsonStr, brand, chargerId);
                    addChargingToDatabase(query, customerSolution);

                }catch(Exception e){
                    e.printStackTrace();
                }

            }
        }

    }

    /**
     *
     * @param jsonStr
     * @param brand
     * @param chargerId
     * @return
     */
    public String makeQuery(String jsonStr, String brand, String chargerId) {

        String query = "INSERT INTO charging (chargerId, date, brand, price, kwh) VALUES ";

        try {
        JSONArray jsonChargingData = new JSONArray(jsonStr);

        for(int k = 0; k < jsonChargingData.length(); k++) {

            JSONObject ChargingObj = jsonChargingData.getJSONObject(k);
            String dateCharging = ChargingObj.getString("date"); //Gets date

            //Formats date to yyyy-mm-dd hh:mm:ss
            dateCharging = dateCharging.substring(0, dateCharging.length() - 5);
            dateCharging = dateCharging.replace("T", " ");

            Float price = ChargingObj.getFloat("nonSmartCostSum"); //Gets price of charging

            Float kwh = ChargingObj.getFloat("kwhSum"); //Gets total kwh used

            query += "('" + chargerId + "', '" + dateCharging + "','" + brand + "', " + price + ", " + kwh + "),";

            if(k == jsonChargingData.length() - 1) {
                query = query.substring(0, query.length() - 1); //Fjerner siste komma
            }
        }


        return query;
        }catch(Exception e){
            e.printStackTrace();
            return null;
        }
    }

    /**
     *
     * @param query
     * @param customerSolution
     */
    private void addChargingToDatabase(String query, String customerSolution){

        try {
            var sql = ClassFactory.getSqlFactory(customerSolution);
            Connection c = sql.getConnection();

            SQLFactory.executeQuery(c, query);


        } catch(SQLException e) {
            System.out.println("Could not find " + customerSolution + " with SqlFactory");

        }

    }


    /**
     *
     * @param bytes
     * @return
     */
    private static String toHexString(byte[] bytes) {
        Formatter formatter = new Formatter();

        for (byte b : bytes) {
            formatter.format("%02x", b);
        }

        return formatter.toString();
    }

    /**
     *
     * @param data
     * @param key
     * @return
     * @throws SignatureException
     * @throws NoSuchAlgorithmException
     * @throws InvalidKeyException
     */
    public static String calculateRFC2104HMAC(String data, String key)
            throws SignatureException, NoSuchAlgorithmException, InvalidKeyException {
        SecretKeySpec signingKey = new SecretKeySpec(key.getBytes(), HMAC_SHA1_ALGORITHM);
        Mac mac = Mac.getInstance(HMAC_SHA1_ALGORITHM);
        mac.init(signingKey);
        return toHexString(mac.doFinal(data.getBytes()));
    }
}
